#include "graphics.h"
#include "JAGUAR.H"
#include "tilemap.h"
#include "game.h"


#define TUNNEL_OFFSET 0
#define INFOBAR_OFFSET 1
#define SHIP_OFFSET 2
#define N_INGAME_BITMAPS 3

#define INFOWINDOW_OFFSET 3
#define N_OUTGAME_BITMAPS 4

#define BESTTIME_DISPLAY_X 176
#define TIME_DISPLAY_X 112
#define SPEED_DISPLAY_X 0

#define TUNNELCENTER 64

#define CODE_1 4893
#define CODE_2 7129
#define CODE_3 6925
#define CODE_4 1156

int nbUnlockedTrack = 9;

/* ---------------------------------------------------------------- */

void GameLoop(int sel)
{
  /* ------------------------------ */
  /*    Game variables              */

  /* Object lists */
static	CBitmap ingamebitmaps[N_INGAME_BITMAPS];
static	CBitmap outgamebitmaps[N_OUTGAME_BITMAPS];
static	CBitmapArray ingamearray;
static	CBitmapArray outgamearray;
static	char listbuff[DLSIZE(N_INGAME_BITMAPS)+32+16];
static	char listbuff2[DLSIZE(N_OUTGAME_BITMAPS)+32+16];
static  char* ptlist;
static  char* ptlist2;
static  CAnim killAnim;
    
  /* Gameplay constants (to adjust for PAL and NTSC systems) */  
  int KDANGLE;
  int KPLAYER_ACCELERATION;
  int KPLAYER_NITRO_ACCELERATION;
  int KPLAYER_TRACK_MAX_SPEED;
  int KPLAYER_OUT_MAX_SPEED;
  int KPLAYER_NITRO_MAX_SPEED;
  int KPLAYER_SLOW_MAX_SPEED;
  int KPLAYER_SLOWDOWN;
  int KPLAYER_SLOW_SLOWDOWN;
  int KPLAYER_VSPEEDF;
  int KPLAYER_FALLSPEEDF;
  int KBASE_FREQ;
  
  /* Player info */
  int player_speedf, player_maxspeedf, player_accelerationf, player_slowdownf;
  int player_altf, player_vspeedf;
  int player_banking = BANK_CENTER;
  int player_accelerating = 0;
  int player_override_rotation = 0;  
  int delta_angle = 0;
  int player_max_rot_speed = 0;
  int player_rot_acc = LOW_INERTIA;
  
  EPlayerState player_state;
  
  /* Sound info */
  ESoundPanning panning = EPanCenter;
  int volleft = VOL_PAN_CENTER; 
  int volright = VOL_PAN_CENTER;
  
  /* Scrolling player and position */  
  int line,scroll_line,angle;
  int posx,posy;
  unsigned int tileline, tracklen;
  char* ptrtrack,*ptrtileset;
  
  int i,j;
  
  /* Timer, game status */
  int tickcount,frames;
  int timer_started;
  int timer_start;
  int endgame;  
  int stats_done = 0;
  char* timedisplay = "xx:xx:x";
  char* speeddisplay = "Speed: xxx kph";
  char* pressfiretxt = "Press fire";
  int decirefreshrate;  
  CTrackInfo* trackinfo;
  int tile_selected = 0;
  int new_tile_selected = 0;
    
  /* Buffers */  
  unsigned char* ptlogic, *ptphysic;
  
  /* Check if track selection is valid */
  if((sel<0)||(sel>=tracks_number)) return;
  
  /* ------------------------------ */
  /*    Init games variables        */
    
  /* Pointers to object lists */    
	ptlist = ALIGN(listbuff);
	ptlist2 = ALIGN(listbuff2);
	
	/* Clear buffers */
	Memset64(&vidmem, 0, 0, 160*100);
  Memset64(&tunnelmem, 0, 0, FBW*FBH);
  Memset64(&tunnelmem2, 0, 0, FBW*FBH);	
	Memset64(&scrollmap, 0x8F448F44, 0x8F448F44, 256*256);

	/* Build object list for before and after game display */
#pragma dontwarn 80	
  MakeBitmap8b(&outgamebitmaps[TUNNEL_OFFSET], &tunnelmem+8*FBW,   0, 8, FBW, FBH-8, 0);
  MakeBitmap8b(&outgamebitmaps[INFOBAR_OFFSET], &tunnelmem,  0, 0, FBW, 8, 0); 
  MakeBitmap(&outgamebitmaps[SHIP_OFFSET], &shipthrustoffbitmap+(BANK_CENTER*PLAYER_W*PLAYER_SKIP_H),(FBW-PLAYER_W)/2,(FBH-PLAYER_YPOS),PLAYER_W,PLAYER_H,O_TRANS);
#pragma popwarn 
	MakeBitmap8b(&outgamebitmaps[INFOWINDOW_OFFSET], &vidmem, (FBW-160)/2, ((FBH-60)>>1), 160, 60, O_TRANS);
	InitBitmapArray(&outgamearray, ptlist2, &outgamebitmaps[0], N_OUTGAME_BITMAPS);	
		
	/* Build object list for in game display */
#pragma dontwarn 80	
  MakeBitmap8b(&ingamebitmaps[TUNNEL_OFFSET], &tunnelmem+8*FBW,   0, 8, FBW, FBH-8, 0);
  MakeBitmap8b(&ingamebitmaps[INFOBAR_OFFSET], &tunnelmem,  0, 0, FBW, 8, 0); 
  MakeBitmap(&ingamebitmaps[SHIP_OFFSET], &shipthrustoffbitmap+(BANK_CENTER*PLAYER_W*PLAYER_SKIP_H),(FBW-PLAYER_W)/2,(FBH-PLAYER_YPOS),PLAYER_W,PLAYER_H,O_TRANS);
#pragma popwarn 
	InitBitmapArray(&ingamearray, ptlist, &ingamebitmaps[0], N_INGAME_BITMAPS);	
	ActiveBitmapArray = &outgamearray;

	LibDrawInit(&vidmem, PIXEL8|WID160, 60);
	SelectFont(&font8x8_8b, 8, 8);
	Memset64(&vidmem,0,0,160*60);
	DrawText(((160-(strlen(pressfiretxt)<<3))>>1) & 0xF8 , 0,pressfiretxt);  
  	
	LibDrawInit(&tunnelmem, PIXEL8|WID320, 200);
	SetBlitMode(XADDPHR);
	SelectFont(&font8x8_8b, 8, 8);
		   			   	
  /* Get track info */  
  trackinfo = &tracks+sel;
  
#pragma dontwarn 80	
  ptrtrack = trackinfo->mPtrTrack+BMP_HEADER_SIZE;  
  ptrtileset = trackinfo->mTileSet+PALETTE_SIZE;
#pragma popwarn
  tracklen = DivU(((int)trackinfo->mPtrEndTrack-(int)ptrtrack), N_TILES_X);
  DrawText(FBW-10*8,0,trackinfo->mName);
  
  TRACE("Init scrollmap\n");	
	LoadCLUT(trackinfo->mTileSet, 256);
	LoadCLUT(&menutiles, 8);
#pragma dontwarn 80
	InitScrollMap(&scrollmap, ptrtileset, ptrtrack);
#pragma popwarn	

  /* PAL and NTSC constants */
  if(systemtype)
	{	
	 decirefreshrate = 6;  /* NTSC */	 
   KDANGLE = 6;
   KPLAYER_ACCELERATION = PLAYER_ACCELERATION;
   KPLAYER_NITRO_ACCELERATION = PLAYER_NITRO_ACCELERATION;
   KPLAYER_TRACK_MAX_SPEED = PLAYER_TRACK_MAX_SPEED;
   KPLAYER_OUT_MAX_SPEED = PLAYER_OUT_MAX_SPEED;
   KPLAYER_NITRO_MAX_SPEED = PLAYER_NITRO_MAX_SPEED;
   KPLAYER_SLOW_MAX_SPEED = PLAYER_SLOW_MAX_SPEED;
   KPLAYER_SLOWDOWN = PLAYER_SLOWDOWN;
   KPLAYER_SLOW_SLOWDOWN = PLAYER_SLOW_SLOWDOWN;
   KPLAYER_VSPEEDF = PLAYER_VSPEEDF;
   KPLAYER_FALLSPEEDF = PLAYER_FALLSPEEDF;
   KBASE_FREQ = BASE_FREQ;   	 
	}
	else
	{
	 decirefreshrate = 5;  /* PAL */	 
   KDANGLE = 7;
   KPLAYER_ACCELERATION = PAL_PLAYER_ACCELERATION;
   KPLAYER_NITRO_ACCELERATION = PAL_PLAYER_NITRO_ACCELERATION;
   KPLAYER_TRACK_MAX_SPEED = PAL_PLAYER_TRACK_MAX_SPEED;
   KPLAYER_OUT_MAX_SPEED = PAL_PLAYER_OUT_MAX_SPEED;
   KPLAYER_NITRO_MAX_SPEED = PAL_PLAYER_NITRO_MAX_SPEED;
   KPLAYER_SLOW_MAX_SPEED = PAL_PLAYER_SLOW_MAX_SPEED;
   KPLAYER_SLOWDOWN = PAL_PLAYER_SLOWDOWN;
   KPLAYER_SLOW_SLOWDOWN = PAL_PLAYER_SLOW_SLOWDOWN;
   KPLAYER_VSPEEDF = PAL_PLAYER_VSPEEDF;
   KPLAYER_FALLSPEEDF = PAL_PLAYER_FALLSPEEDF;
   KBASE_FREQ = BASE_FREQ_PAL;		 
  }  
    
  /* Player */		   
	player_speedf = 0;	
  player_maxspeedf = KPLAYER_TRACK_MAX_SPEED;
  player_accelerationf = KPLAYER_ACCELERATION;
	player_slowdownf = KPLAYER_SLOWDOWN; 
	player_altf = 0;
  player_vspeedf = 0;
	player_max_rot_speed = KDANGLE;
	
	
	/* Scrolling variables */
	line = 0;
	scroll_line = 0;
	angle = 0;
	tileline = 7;
  gpuscrolltileline = 7;
  
  	/* Chrono */	
	timer_started = 0;
  timer_start=0;
	player_state = EIdle;	
	endgame = 0;
	frames = 0;
	timedisplay = "00:00:0";	
    
  /* Refresh time display buffer, otherwhise we get values from previous round */ 
  timedisplay[0] = '0';
  timedisplay[1] = '0';
  timedisplay[3] = '0';
  timedisplay[4] = '0';
  timedisplay[6] = '0';  
	DrawText(TIME_DISPLAY_X,0,timedisplay);

  {
  int tenthsecondmod =0;        
  char digitsbuf[4];
  int minutes = DivUmod(trackinfo->mBestTime,600, &tenthsecondmod);
  ConvertDec(tenthsecondmod,digitsbuf,3);
  timedisplay[3] = digitsbuf[0]; 
  timedisplay[4] = digitsbuf[1];
  timedisplay[6] = digitsbuf[2];
  
  ConvertDec(minutes,digitsbuf,2);
  timedisplay[0] = digitsbuf[0]; 
  timedisplay[1] = digitsbuf[1];
  DrawText(BESTTIME_DISPLAY_X,0,timedisplay);
  }
  
  if(systemtype)  
  {
    speeddisplay = "Speed:000mph";  /* NTSC */
  }
  else
  {
    speeddisplay = "Speed:000kph"; /* PAL */
  }  
  DrawText(SPEED_DISPLAY_X,0,speeddisplay);
  
  
  /* Start sound */
  SetVol(4, VOL_PAN_CENTER);
  SetVol(5, VOL_PAN_CENTER);
  SetFreq(4,BASE_FREQ);
  SetFreq(5,BASE_FREQ+SURROUND_FREQ);
  SetFreq(7,16000);
  SetVol(7,60);
  PlaySampleDoubleChan(4, JET_SND);
 
	
	TRACE("Main loop\n");
	  
	ptlogic = &tunnelmem2;
	ptphysic = &tunnelmem;

#ifdef BENCHMARK	
        timer_started = -1;
        timer_start = ticks;
        player_state = EActive;        
        ActiveBitmapArray = &ingamearray;
        tickcount = ticks;     
        frames = 0;   
        PlayModule();
#endif
        		
	/* Main loop */
	while(!endgame)
	{
	  unsigned char* ptrtmp;
	  char digitsbuf[8];
	  int newtileline = 0;
	  	  
		ReadJoypads();
				
		if((joy1cur & (1<<KEY_STAR))
    &&(joy1edge & (1<<KEY_STAR)))
    {     
      TRACE("Close\n")  
      CLOSECONSOLE    
    }
    
    if((joy1cur & (1<<KEY_HASH))
    &&(joy1edge & (1<<KEY_HASH)))
    {
      endgame = -1;
    }

#ifdef BENCHMARK
  /*WaitVBL();*/
#else
  WaitVBL();
#endif
    
		DrawTunnel(&scrollmap + (angle&(SCROLLMAPW-1)) + scroll_line*SCROLLMAPW,
                TUNNELCENTER-(player_altf>>16),
                ptlogic);	

    /* ------------ */
    /* Tile effect */
    switch(tile_selected)
    {     
    case TILE_TRACK:
      player_maxspeedf = KPLAYER_TRACK_MAX_SPEED;
      player_accelerationf = KPLAYER_ACCELERATION;
      player_slowdownf = KPLAYER_SLOWDOWN;
      player_override_rotation = 0;
      player_max_rot_speed = KDANGLE;
      break;
    case TILE_NITRO:
      player_maxspeedf = KPLAYER_NITRO_MAX_SPEED;
      player_accelerationf = KPLAYER_NITRO_ACCELERATION;
      player_slowdownf = 0;    
      player_override_rotation = 0;
      player_max_rot_speed = KDANGLE;
      break;
    case TILE_SLOW:
      player_maxspeedf = KPLAYER_SLOW_MAX_SPEED;
      player_accelerationf = 0;
      player_slowdownf = KPLAYER_SLOW_SLOWDOWN;
      player_override_rotation = 0;
      player_max_rot_speed = KDANGLE;	   
      break;      
    case TILE_BUMP:
      player_vspeedf = KPLAYER_VSPEEDF; 
      tile_selected =  IN_AIR;    
      player_override_rotation = 0;
      player_max_rot_speed = KDANGLE;
	    break;
    case TILE_FIRE:
      player_maxspeedf = KPLAYER_OUT_MAX_SPEED;  
      player_accelerationf = KPLAYER_ACCELERATION;   
      player_slowdownf = KPLAYER_SLOWDOWN;
      player_override_rotation = 0;
      player_max_rot_speed = KDANGLE;	    
      panning = EPanCenter;
      if((player_state != EDestroyed)&&(player_state!=EFinish))
      {
        player_state = EDestroyed;        
#pragma dontwarn 80
        ingamebitmaps[SHIP_OFFSET].addr = &shipkillbitmap;
#pragma popwarn
        MakeAnim(&killAnim, &ingamebitmaps[SHIP_OFFSET], PLAYER_W*PLAYER_SKIP_H*2, 6, 1);
        killAnim.loopStep = 2;
        PlaySampleDoubleChan(4, CRASH_SND);
      }
      break;
    case TILE_RIGHT:
      player_maxspeedf = KPLAYER_TRACK_MAX_SPEED;
      player_accelerationf = KPLAYER_ACCELERATION;
      player_slowdownf = KPLAYER_SLOWDOWN;
      panning = EPanRight;
      player_override_rotation = -1;
      player_max_rot_speed = KDANGLE*2;	    
      break;
    case TILE_LEFT:
      player_maxspeedf = KPLAYER_TRACK_MAX_SPEED;
      player_accelerationf = KPLAYER_ACCELERATION;
      player_slowdownf = KPLAYER_SLOWDOWN;
      panning = EPanLeft;
      player_override_rotation = -1;
      player_max_rot_speed = KDANGLE*2;	    
      break;
    case TILE_INERTIA:
      player_maxspeedf = KPLAYER_TRACK_MAX_SPEED;
      player_accelerationf = KPLAYER_ACCELERATION;
      player_slowdownf = KPLAYER_SLOWDOWN;
      player_override_rotation = 0;
      player_max_rot_speed = KDANGLE;
      break;
       
    case IN_AIR:
      player_slowdownf = 0;
      player_override_rotation = 0;
      player_max_rot_speed = KDANGLE;
      break;
          
    default:  
      player_maxspeedf = KPLAYER_OUT_MAX_SPEED;  
      player_accelerationf = KPLAYER_ACCELERATION;   
      player_slowdownf = KPLAYER_SLOWDOWN; 
      player_override_rotation = 0;
      player_max_rot_speed = KDANGLE;            
    }
    
    /* -------------------------------- */
    /* Player control                   */
    switch(player_state)
    {
    case EIdle:  
      panning = EPanCenter;    
      if((joy1cur & (1<<FIRE_B)) && (joy1edge & (1<<FIRE_B)))
      {
        timer_started = -1;
        timer_start = ticks;
        player_state = EActive;
        PlayModule();
        WaitVBL();        
        ActiveBitmapArray = &ingamearray;
        tickcount = ticks;     
        frames = 0;   
      }
      break;
    case EActive:
      if(!player_override_rotation)
      { 
        if(joy1cur & (1<<JOY_LEFT))
        {
          panning = EPanLeft;       
        }
        else if(joy1cur & (1<<JOY_RIGHT))
        {
          panning = EPanRight;                   
        }
        else
        {
          panning = EPanCenter;        
        }
      }
      
      if((joy1cur & (1<<FIRE_B)) || (tile_selected == TILE_NITRO))
      {
        if(player_speedf<=player_maxspeedf)
        {
          player_speedf += player_accelerationf;
          if(player_speedf>player_maxspeedf) player_speedf= player_maxspeedf;
        }  
        player_accelerating = -1; 
      }
      else
      {
        player_accelerating = 0;
      }
      break;
    case EFinish:
      if((joy1cur & (1<<FIRE_B)) && (joy1edge & (1<<FIRE_B)))
      {
        endgame = -1;
        panning = EPanCenter;
        player_accelerating = 0;
      }
      break;
    case EDestroyed:
      panning = EPanCenter;
      if((joy1cur & (1<<FIRE_B)) && (joy1edge & (1<<FIRE_B)))
      {
        int safeland = 0;
        
        player_state = EActive;
        player_speedf = 0;
        player_maxspeedf = KPLAYER_TRACK_MAX_SPEED;
        player_accelerationf = KPLAYER_ACCELERATION;
      	player_slowdownf = KPLAYER_SLOWDOWN; 
      	player_altf = 0;
        player_vspeedf = 0;        

        do 
        {
          int testx, testy, testresult = 0;
          line &= ~(TILE_H-1);
          line -= 6*TILE_H;        
          if(line < 0)
          {
           line = 0;           
          }
          scroll_line = 0;
          tileline = (line >> TILESH_H) +7;   
          testx = (((angle+(TILE_W>>1)) >> TILESH_W) + 3) & (N_TILES_X-1);
  	      testy = ((line+8*TILE_H) >> TILESH_H)-7; 		
          testresult = (int)*(ptrtrack+testx+testy*N_TILES_X);
          if(testresult!= TILE_FIRE)
          {
            safeland = -1;
          }     
        }
        while(!safeland);
        InitScrollMap(&scrollmap, ptrtileset, ptrtrack+N_TILES_X*(tileline-7)); 
        gpuscrolltileline = 7;
        panning = EPanCenter;
        SetVol(4, VOL_PAN_CENTER);
        SetVol(5, VOL_PAN_CENTER);        
        PlaySampleDoubleChan(4, JET_SND);       
      }
      break;
    default:
      ;
    }

    player_speedf -= player_slowdownf;
    if(player_speedf<0) player_speedf=0;  

    player_vspeedf -= KPLAYER_FALLSPEEDF;
  
    player_altf += player_vspeedf;
    if(player_altf < 0)
    {
      player_altf = 0;
      player_vspeedf = 0;
    }
    
    if(panning == EPanLeft)
    {
      delta_angle -= player_rot_acc;
      if(delta_angle < -player_max_rot_speed) delta_angle =  -player_max_rot_speed;
    }
    else if(panning == EPanRight)
    {
      delta_angle += player_rot_acc;
      if(delta_angle > player_max_rot_speed) delta_angle = player_max_rot_speed;
    }
    else
    {
      if(delta_angle > 0) 
      {
        delta_angle -= player_rot_acc;
        if(delta_angle <0) delta_angle = 0;
      }
      if(delta_angle < 0)
      { 
        delta_angle += player_rot_acc;
        if(delta_angle>0) delta_angle = 0;
      }      
    }
    
    angle += delta_angle;
   
#ifdef BENCHMARK
    line += (KPLAYER_NITRO_MAX_SPEED>>16);  
    scroll_line = (scroll_line +   (KPLAYER_NITRO_MAX_SPEED>>16)) & SCROLLMAPH-1;    
    newtileline = (line+8*TILE_H) >> TILESH_H;
#else
    line += (player_speedf>>16);  
    scroll_line = (scroll_line +   (player_speedf>>16)) & SCROLLMAPH-1;    
    newtileline = (line+8*TILE_H) >> TILESH_H;
#endif
		
		/* --------------------- */
    /*  Collision detection */
    if(player_altf == 0)
    {   
  	  posx = (((angle+(TILE_W>>1)) >> TILESH_W) + 3) & (N_TILES_X-1);
  	  posy = newtileline-7;    		
  		
      new_tile_selected = (int)*(ptrtrack+posx+posy*N_TILES_X);    
  		if(new_tile_selected != tile_selected)
  		{
  		  tile_selected = new_tile_selected;      
        if(tile_selected == TILE_INERTIA)
        {
          player_rot_acc = HIGH_INERTIA;
          PlaySample(7, ELECTRIC_SND);
        }
  	  }
  	}
	
    GPUsync();    
    ptrtmp = ptlogic;
    ptlogic = ptphysic;
    ptphysic = ptrtmp;
    
    if(ActiveBitmapArray == &ingamearray)
    {
      ingamebitmaps[TUNNEL_OFFSET].addr = ptphysic+8*FBW;
    }
    else
    {
      outgamebitmaps[TUNNEL_OFFSET].addr = ptphysic+8*FBW;
    } 
   
    /* Speed display, 
    Sprite animation,
      Time display,        
      sound changes */
    if((frames & 0x1) == 0)
    {
      int speed = player_speedf >> 12;      
      DrawDec(SPEED_DISPLAY_X+6*8, 0, speed, 3);

#pragma dontwarn 80
      switch(panning)
      {
      case EPanLeft:
        if(player_banking>BANK_LEFT) player_banking--;
        break;
      case EPanRight:
        if(player_banking<BANK_RIGHT) player_banking++;
        break;      
      case EPanCenter:
      default:
        if(player_banking<BANK_CENTER) player_banking++;
        else if(player_banking>BANK_CENTER) player_banking--;
      }      
      if(player_state != EDestroyed)
      {
        unsigned short* baseptr;
        if(player_accelerating) baseptr = &shipbitmap;
        else baseptr = &shipthrustoffbitmap;
        ingamebitmaps[SHIP_OFFSET].addr = baseptr+(player_banking*PLAYER_W*PLAYER_SKIP_H);
      }
      else
      {
        UpdateAnim(&killAnim, 1);
      }
#pragma popwarn

      if(timer_started && (player_state != EFinish))
  	  {	     
  	     int tenthsecond = DivU(ticks-timer_start, decirefreshrate);
         if((frames & 0x7) == 0)  /* update seconds display every 8 frames */
         {
            int tenthsecondmod;
            int minutes = DivUmod(tenthsecond,600, &tenthsecondmod);
            ConvertDec(tenthsecondmod,digitsbuf,3);
            timedisplay[3] = digitsbuf[0]; 
            timedisplay[4] = digitsbuf[1];
            timedisplay[6] = digitsbuf[2];
            if((frames & 0xF) == 0) /* update minutes display every 16 frames */
            {
              ConvertDec(minutes,digitsbuf,2);
              timedisplay[0] = digitsbuf[0]; 
              timedisplay[1] = digitsbuf[1];
              DrawText(TIME_DISPLAY_X,0,timedisplay);
            }
            else
            {
              DrawText(TIME_DISPLAY_X+3*8,0,timedisplay+3);
            }
         }
         else
         {
            ConvertDec(tenthsecond,digitsbuf,1);
            timedisplay[6] = *digitsbuf;
  	        DrawText(TIME_DISPLAY_X+6*8,0,timedisplay+6);
         }
      }

    
      switch(panning)
      {
      case EPanRight:
        if(volright<VOL_PAN_SIDE_HI)
          volright+= 4;
        if(volleft>VOL_PAN_SIDE_LOW)
          volleft-= 4;     
        break;
        
      case EPanLeft:
        if(volleft<VOL_PAN_SIDE_HI)
          volleft+= 4;
        if(volright>VOL_PAN_SIDE_LOW)
          volright-= 4; 
        break;
 
      case EPanCenter: 
      default:      
        if(volleft>VOL_PAN_CENTER) volleft -= 4;
        if(volleft<VOL_PAN_CENTER) volleft += 4;
        if(volright>VOL_PAN_CENTER) volright -= 4;
        if(volright<VOL_PAN_CENTER) volright += 4;        
      }
      SetVol(4,volleft-(player_altf>>18));
      SetVol(5,volright-(player_altf>>18));
      SetFreq(4,KBASE_FREQ+(player_speedf>>6));
      SetFreq(5,KBASE_FREQ+SURROUND_FREQ+(player_speedf>>6));
      ApplyFreqVolChanges();
    }
    
    if((newtileline>=(tracklen-9))&&(player_state != EFinish))
    { 
      if(player_state == EDestroyed)
      {
        int safeland = 0;
        
        player_state = EActive;
        player_speedf = 0;
        player_maxspeedf = KPLAYER_TRACK_MAX_SPEED;
        player_accelerationf = KPLAYER_ACCELERATION;
      	player_slowdownf = KPLAYER_SLOWDOWN; 
      	player_altf = 0;
        player_vspeedf = 0;        

        do 
        {
          int testx, testy, testresult = 0;
          line &= ~(TILE_H-1);
          line -= 8*TILE_H;        
          if(line < 0)
          {
           line = 0;           
          }
          scroll_line = 0;
          tileline = (line >> TILESH_H) +7;   
          testx = (((angle+(TILE_W>>1)) >> TILESH_W) + 3) & (N_TILES_X-1);
  	      testy = ((line+8*TILE_H) >> TILESH_H)-7; 		
          testresult = (int)*(ptrtrack+testx+testy*N_TILES_X);
          if(testresult!= TILE_FIRE)
          {
            safeland = -1;
          }     
        }
        while(!safeland);
        InitScrollMap(&scrollmap, ptrtileset, ptrtrack+N_TILES_X*(tileline-7)); 
        gpuscrolltileline = 7;
        panning = EPanCenter;
        SetVol(4, VOL_PAN_CENTER);
        SetVol(5, VOL_PAN_CENTER);        
        PlaySampleDoubleChan(4, JET_SND);         
      }
      else
      {      
        char* finishtxt = "Finish";
        char* bestttimetxt = "Best time!";
        int tenthsecondmod;    
        int player_time = ticks-timer_start;      
        int tenthsecond = DivU(ticks-timer_start, decirefreshrate);            
        int minutes = DivUmod(tenthsecond,600, &tenthsecondmod);
        ConvertDec(tenthsecondmod,digitsbuf,3);
        timedisplay[3] = digitsbuf[0]; 
        timedisplay[4] = digitsbuf[1];
        timedisplay[6] = digitsbuf[2];
        
        ConvertDec(minutes,digitsbuf,2);
        timedisplay[0] = digitsbuf[0]; 
        timedisplay[1] = digitsbuf[1];
        DrawText(TIME_DISPLAY_X,0,timedisplay);
  	    
  	    LibDrawInit(&vidmem, PIXEL8|WID160, 60);
      	SelectFont(&font8x8_8b, 8, 8);
      	Memset64(&vidmem,0,0,160*60);
      	DrawText((160-(strlen(finishtxt)<<3))>>1 & 0xF8,0,finishtxt);
        /* Best time ? */
        if(tenthsecond<trackinfo->mBestTime)
        {
          trackinfo->mBestTime = tenthsecond;          
          DrawText((160-(strlen(bestttimetxt)<<3))>>1 & 0xF8,8,bestttimetxt);
          if(trackinfo->mCode != 0)
          {
            DrawText(5*8,16,"Code:");
            DrawDec(11*8,16,trackinfo->mCode,4);
          }
          trackinfo->mUnlocked = -1;
        } 	
        LibDrawInit(&tunnelmem, PIXEL8|WID320, 200);
        SelectFont(&font8x8_8b, 8, 8);
        player_state = EFinish;      
        panning = EPanCenter;  
            
        WaitVBL();
        ActiveBitmapArray = &outgamearray;
        ptrtmp = ptlogic;
        ptlogic = ptphysic;
        ptphysic = ptrtmp;
        outgamebitmaps[TUNNEL_OFFSET].addr = ptphysic+8*FBW;
      }
    }
	   
    /* --------------- */
    /* Scrolling */		
		if(newtileline > tileline)
		{	  
		  int scantileline;		  
		  tileline = newtileline;
		 
		  if(tileline >= tracklen)
		  {
		    scantileline = (tileline & (((SCROLLMAPH*2)/TILE_H)-1)) + tracklen - ((SCROLLMAPH*2)/TILE_H);		
      }
      else
      {
        scantileline = tileline;
      }

	    DrawTileLine(ptrtrack+N_TILES_X*scantileline, ptrtileset); 
      GPUsync();
    }    
            
    frames++;

#ifdef BENCHMARK			   
    if((frames == 192)&&(timer_started))
    {   
      if(!stats_done)
      {
        tickcount = ticks-tickcount;        
        DrawDec(256,0,tickcount,3); 
                
        TRACE("\nTicks: ")  
        TRACEDEC(tickcount,8)        
        TRACE("\nFrames: ")  
        TRACEDEC(frames,8)
        TRACE("\n")
        
        stats_done = -1;
      }
    } 
#endif   
	}
}


/* ---------------------------------------------------------------- */

#define TITLE_OFFSET 2
#define TRACK_INFO_OFFSET 3
#define N_MENU_BITMAPS 4

#define TITLE_W 232
#define TITLE_H 66
#define TRACKINFO_W 256
#define TRACKINFO_H 10

void DrawTrackInfo(int sel)
{
  char digitsbuf[4];
  int minutes,tenthseconds;
  char* timedisplay = ":xx:x";
  CTrackInfo* trackArray = &tracks;
  if((sel < 0)||(sel>=tracks_number)) return;
  
      
  Memset64(&vidmem, 0, 0, TRACKINFO_W*TRACKINFO_H);
  LibDrawInit(&vidmem, PIXEL8|WID256, 10);
	SetBlitMode(XADDPHR);
	SelectFont(&font8x8_8b, 8, 8);
	
	if(sel>=nbUnlockedTrack)
	{
    DrawDec(8,0,sel+1,2);
  	DrawText(24,0,"/");
  	DrawDec(32,0,tracks_number,2);
    DrawText(56,0,"LOCKED");	  
  }
  else
	{
  	DrawDec(8,0,sel+1,2);
  	DrawText(24,0,"/");
  	DrawDec(32,0,tracks_number,2);
    DrawText(56,0,trackArray[sel].mName);
    minutes = DivUmod(trackArray[sel].mBestTime, 600, &tenthseconds);
    ConvertDec(tenthseconds,digitsbuf,3);
    timedisplay[1] = digitsbuf[0];
    timedisplay[2] = digitsbuf[1];
    timedisplay[4] = digitsbuf[2];
    DrawDec(256-(8*8),0,minutes,2);   
    DrawText(256-(6*8),0,timedisplay);
  }
}

void DrawEnterCode(int num)
{
  Memset64(&vidmem, 0, 0, TRACKINFO_W*TRACKINFO_H);
  LibDrawInit(&vidmem, PIXEL8|WID256, 10);
	SetBlitMode(XADDPHR);
	SelectFont(&font8x8_8b, 8, 8);
	
	DrawText(8,0,"Enter code:");
	DrawDec(13*8,0,num,4);
}

static char* ptrmenutiles = &menutiles;

#define NB_MESSAGES 5
static char* messages[NB_MESSAGES];
                            
int MenuLoop()
{
typedef enum
{
  ETrackSelection = 0,
  EEnterCode
} menu_mode;

static CBitmap menuBitmaps[N_MENU_BITMAPS];
static CBitmapArray menuArray;
static char listbuff[DLSIZE(N_MENU_BITMAPS)+32+16];
static char* ptlist;
  /* Buffers */  
  unsigned char* ptlogic, *ptphysic;
  int scroll_line = 0;
  int selection = -1;
  int maskoffset;
  int tickmessage = 0;
  int tickcount = 0;
  int messagecounter = 0;
  int frames = 0;
  int code = 0;
  CTrackInfo* trackArray = &tracks;
  menu_mode mode = ETrackSelection;
static int choice = 0;
    
  TRACE("\nMenuLoop\n")
  messages[0] = "Code by DrTypo";
  messages[1] = "Graphics by Marcus Losbjer";
  messages[2] = "Music by Marcus Losbjer";
  messages[3] = "Beta test by JagChris";
  messages[4] = "U-235 Sound engine by LinkoVitch";
  
  trackArray[8].mCode = CODE_1;
  trackArray[11].mCode = CODE_2;
  trackArray[14].mCode = CODE_3;
  
  Memset64(&tunnelmem, 0, 0, FBW*FBH);
  Memset64(&tunnelmem2, 0, 0, FBW*FBH);
  Memset64(&scrollmap, 0x8F448F44, 0x8F448F44, 256*256);
  Memset64(&vidmem, 0, 0, 160*100);
  ptlist = ALIGN(listbuff);
    
  /* Build object list for in menu display */
#pragma dontwarn 80	
  MakeBitmap8b(&menuBitmaps[TUNNEL_OFFSET], &tunnelmem+8*FBW,   0, 8, FBW, FBH-8, 0);
  MakeBitmap8b(&menuBitmaps[INFOBAR_OFFSET], &tunnelmem,  0, 0, FBW, 8, 0);
  MakeBitmap(&menuBitmaps[TITLE_OFFSET], &titlebitmap,(FBW-TITLE_W)/2,50,TITLE_W,TITLE_H,O_TRANS);
  MakeBitmap8b(&menuBitmaps[TRACK_INFO_OFFSET], &vidmem, (FBW-TRACKINFO_W)/2,150, TRACKINFO_W,TRACKINFO_H, O_TRANS);
#pragma popwarn    
	InitBitmapArray(&menuArray, ptlist, &menuBitmaps[0], N_MENU_BITMAPS);  
	ActiveBitmapArray = &menuArray;

  /* Init scrollmap */
	LoadCLUT(ptrmenutiles, 256);
#pragma dontwarn 80  
	InitScrollMap(&scrollmap, ptrmenutiles+PALETTE_SIZE, &menumap+BMP_HEADER_SIZE);
	InitScrollMap(&scrollmap+SCROLLMAPW*SCROLLMAPH, ptrmenutiles+PALETTE_SIZE, &menumap+BMP_HEADER_SIZE);
#pragma popwarn	

  DrawTrackInfo(choice);

	LibDrawInit(&tunnelmem, PIXEL8|WID320, 200);
	SetBlitMode(XADDPHR);
	SelectFont(&font8x8_8b, 8, 8);
	DrawText(((FBW-strlen(messages[0])*8)>>1)&0xFFF8,0, messages[0]);
	
	ptlogic = &tunnelmem2;
	ptphysic = &tunnelmem;

  SetVol(4, 0);
  SetVol(5, 0);
    
  tickmessage = ticks;
  tickcount = ticks;
  frames = 0;
  while(selection == -1)
  {
     unsigned char* ptrtmp;
      
    ReadJoypads();
    
    if(  ((joy1cur & (1<<JOY_UP)) && (joy1edge & (1<<JOY_UP)))
       ||((joy1cur & (1<<JOY_DOWN)) && (joy1edge & (1<<JOY_DOWN)))
      )
    {
      mode = ((mode+1)&0x1);
      if(mode == 0)
        DrawTrackInfo(choice);
    }
    
    if(mode == ETrackSelection)
    {
      if((joy1cur & (1<<FIRE_B)) && (joy1edge & (1<<FIRE_B)))
      {           
        selection = choice;   
      }
      if((joy1cur & (1<<JOY_LEFT)) && (joy1edge & (1<<JOY_LEFT)))
      {           
        choice--;
        if(choice<0) choice = tracks_number-1; 
        DrawTrackInfo(choice);  
      }
      if((joy1cur & (1<<JOY_RIGHT)) && (joy1edge & (1<<JOY_RIGHT)))
      {           
        choice++;
        if(choice>=tracks_number) choice = 0; 
        DrawTrackInfo(choice);  
      }
    }
    else
    {      
      if((joy1cur & (1<<KEY_0)) && (joy1edge & (1<<KEY_0)))
        code = code*10;
      if((joy1cur & (1<<KEY_1)) && (joy1edge & (1<<KEY_1)))
        code = code*10 + 1;
      if((joy1cur & (1<<KEY_2)) && (joy1edge & (1<<KEY_2)))
        code = code*10 + 2;  
      if((joy1cur & (1<<KEY_3)) && (joy1edge & (1<<KEY_3)))
        code = code*10 + 3;
      if((joy1cur & (1<<KEY_4)) && (joy1edge & (1<<KEY_4)))
        code = code*10 + 4;
      if((joy1cur & (1<<KEY_5)) && (joy1edge & (1<<KEY_5)))
        code = code*10 + 5;
      if((joy1cur & (1<<KEY_6)) && (joy1edge & (1<<KEY_6)))
        code = code*10 + 6;
      if((joy1cur & (1<<KEY_7)) && (joy1edge & (1<<KEY_7)))
        code = code*10 + 7;
      if((joy1cur & (1<<KEY_8)) && (joy1edge & (1<<KEY_8)))
        code = code*10 + 8;
      if((joy1cur & (1<<KEY_9)) && (joy1edge & (1<<KEY_9)))
        code = code*10 + 9;  
      if(code > 9999) code = 0;
            
      DrawEnterCode(code);
      
      if((code == CODE_1)&&(nbUnlockedTrack<12))
      {
        code = 0;
        nbUnlockedTrack = 12;
        choice = 9;
        mode = ETrackSelection;
        DrawTrackInfo(choice);        
      }  
      else if((code == CODE_2)&&(nbUnlockedTrack<15))
      {
        code = 0;
        nbUnlockedTrack = 15;
        choice = 12;
        mode = ETrackSelection;
        DrawTrackInfo(choice);
      }
      else if((code == CODE_3)&&(nbUnlockedTrack<18))
      {
        code = 0;
        nbUnlockedTrack = 18;
        choice = 15;
        mode = ETrackSelection;
        DrawTrackInfo(choice);
      }
      else if(code == CODE_4)
      {
        code = 0;
        mode = ETrackSelection;
        selection = 42;
      }
      else 
      {
        DrawEnterCode(code);
      }               
    }
    		
    #ifdef BENCHMARK
    /*WaitVBL();*/
    #else
    WaitVBL();
    #endif
    DrawTunnel(&scrollmap + scroll_line*SCROLLMAPW,
              TUNNELCENTER,
              ptlogic);
    scroll_line = (scroll_line + 4) & SCROLLMAPH-1;    
    
    GPUsync();    
    
    ApplyFreqVolChanges();
    
    ptrtmp = ptlogic;
    ptlogic = ptphysic;
    ptphysic = ptrtmp;      
    menuBitmaps[TUNNEL_OFFSET].addr = ptphysic+8*FBW;
    
#ifndef BENCHMARK
    if((ticks -  tickmessage) > 300)
    {
      messagecounter++;
      tickmessage = ticks;
      if(messagecounter>= NB_MESSAGES)
      {
        messagecounter = 0;          
      }
     	LibDrawInit(&tunnelmem, PIXEL8|WID320, 200);	
      SelectFont(&font8x8_8b, 8, 8);
      Memset64(&tunnelmem,0,0,FBW*8);
      DrawText(((FBW-strlen(messages[messagecounter])*8)>>1)&0xFFF8,0, messages[messagecounter]);
    }
#endif
   
#ifdef BENCHMARK    
    frames++;
       
    if(frames == 192)
    {   
     	LibDrawInit(&tunnelmem, PIXEL8|WID320, 200);	
      SelectFont(&font8x8_8b, 8, 8);    
      tickcount = ticks-tickcount;        
      DrawDec(256,0,tickcount,3); 
              
      TRACE("\nTicks: ")  
      TRACEDEC(tickcount,8)        
      TRACE("\nFrames: ")  
      TRACEDEC(frames,8)
      TRACE("\n")     
    }   
#endif                 
  }
  
  return selection;
}

#define N_ENDING_BITMAPS 2

#define END_BITMAP_OFFSET 0
#define TEXT_BITMAP_OFFSET 1
#define TEXT_AREA_W 160
#define TEXT_AREA_H 100

void Outtro()
{
static CBitmap endBitmaps[N_ENDING_BITMAPS];
static CBitmapArray endBitmapsArray;
static char listbuff[DLSIZE(N_ENDING_BITMAPS)+32+16];
static char* ptlist;

  /* Buffers */
  int tickcount = 0;
  int backtomenu = 0;
  
  TRACE("\nOuttro\n")

  Memset64(&vidmem, 0, 0, 160*100);
  ptlist = ALIGN(listbuff);
    
  /* Build object list for in menu display */
#pragma dontwarn 80	
  MakeBitmap(&endBitmaps[END_BITMAP_OFFSET], &endbitmap,0,0,FBW,FBH, 0);
  MakeBitmap8b(&endBitmaps[TEXT_BITMAP_OFFSET], &vidmem, (FBW-TEXT_AREA_W)>>1,(FBH-TEXT_AREA_H)>>1, TEXT_AREA_W,TEXT_AREA_H, O_TRANS);
#pragma popwarn    
	InitBitmapArray(&endBitmapsArray, ptlist, &endBitmaps[0], N_ENDING_BITMAPS);  
	ActiveBitmapArray = &endBitmapsArray;

  /* Init scrollmap */
/*	LoadCLUT(ptrmenutiles, 256);
	LibDrawInit(&vidmem, PIXEL8|WID160, 100);
	SetBlitMode(XADDPHR);
	SelectFont(&font8x8_8b, 8, 8);
	DrawText(((FBW-strlen(messages[0])*8)>>1)&0xFFF8,0, messages[0]);*/

  tickcount = ticks;

  while(!backtomenu)
  { 
    WaitVBL();  
    ReadJoypads();
        
    if((joy1cur & (1<<FIRE_B)) && (joy1edge & (1<<FIRE_B)))
    {           
      backtomenu = -1;   
    }
    
                  
  }  
}