
#ifndef __GAME_H__
#define __GAME_H__

/* --------------------- */
/* Game constants       */

/*#define BENCHMARK*/

/* Generics constants */
#define BMP_HEADER_SIZE 1078
#define PALETTE_SIZE 512
#define FBW 320
#define FBH_NTSC 200
#define FBH_PAL 232

/* Player */
#define PLAYER_YPOS 32
#define PLAYER_W 56
#define PLAYER_H 26
#define PLAYER_SKIP_H 32
#define BANK_LEFT 0
#define BANK_CENTER 2
#define BANK_RIGHT 4

/* Tiles */
#define TILE_OUT  0x00
#define TILE_TRACK 0x01
#define TILE_NITRO 0x02
#define TILE_SLOW 0x03
#define TILE_BUMP 0x04
#define TILE_FIRE 0x05
#define TILE_RIGHT 0x06
#define TILE_LEFT 0x07
#define TILE_INERTIA 0x08
#define IN_AIR    0x0F

/* Tuning */
#define LOW_INERTIA 3
#define HIGH_INERTIA 1


/* NTSC values */
#define PLAYER_ACCELERATION 0x6000
#define PLAYER_NITRO_ACCELERATION 0xA000

#define PLAYER_TRACK_MAX_SPEED    0x84000
#define PLAYER_OUT_MAX_SPEED    0x44000
#define PLAYER_NITRO_MAX_SPEED    0xE0000
#define PLAYER_SLOW_MAX_SPEED 0x22000

#define PLAYER_SLOWDOWN         0x4000
#define PLAYER_SLOW_SLOWDOWN   0xA000

#define PLAYER_VSPEEDF     0x6E000
#define PLAYER_FALLSPEEDF  0x06800

/* PAL values */
#define PAL_PLAYER_ACCELERATION 0x7200
#define PAL_PLAYER_NITRO_ACCELERATION 0xC000

#define PAL_PLAYER_TRACK_MAX_SPEED    0xA5000
#define PAL_PLAYER_OUT_MAX_SPEED    0x55000
#define PAL_PLAYER_NITRO_MAX_SPEED    0x110000
#define PAL_PLAYER_SLOW_MAX_SPEED 0x24000

#define PAL_PLAYER_SLOWDOWN         0x5000
#define PAL_PLAYER_SLOW_SLOWDOWN   0xD000

#define PAL_PLAYER_VSPEEDF     0x70000
#define PAL_PLAYER_FALLSPEEDF  0x08000

/* Sounds contants */

/* Indices in sample bank */
#define JET_SND  0
#define CRASH_SND  1
#define JAGUAR_SND 2
#define ELECTRIC_SND 3
#define BLANK_SND 4

/* Sound Parameters */
#define BASE_FREQ 8500
#define BASE_FREQ_PAL 8000
#define SURROUND_FREQ 200

#define VOL_PAN_CENTER 40
#define VOL_PAN_SIDE_HI 64
#define VOL_PAN_SIDE_LOW 20

/* --------------------- */
/* Game types            */
typedef struct
{
  char* mPtrTrack;
  char* mPtrEndTrack;
  char* mTileSet;
  char* mName;
  int   mBestTime;
  int   mCode;
  int   mUnlocked;
} CTrackInfo;

/* Player info */
typedef enum
{
  EIdle =0,
  EActive,
  EFinish,
  EDestroyed
} EPlayerState;

/* Sound panning */
typedef enum
{
  EPanCenter = 0,
  EPanLeft,
  EPanRight
} ESoundPanning;


/* --------------------- */
/* Externals            */

/* Rendering */
extern int systemtype;
extern int FBH;
extern unsigned short vidmem; 

extern unsigned char tunnelmem;
extern unsigned char tunnelmem2;
extern unsigned char* ptrtunnelmem;
extern unsigned char masktunnel;
extern unsigned short lutdata;
extern unsigned char scrollmap;
extern unsigned int gpuscrolltileline;

/* Bitmaps */
extern unsigned char font8x8_8b;
extern unsigned short shipbitmap;	
extern unsigned short shipthrustoffbitmap;
extern unsigned short shipkillbitmap;
extern unsigned short titlebitmap;
extern unsigned char menutiles;
extern unsigned char menumap;
extern unsigned short endbitmap;

/* Soundtracker module and sounds */
extern unsigned int module;
extern unsigned int ingamemod1;
extern unsigned int ingamemod2;
extern unsigned int ingamemod3;
extern unsigned int endmod;
extern unsigned char jetsound;
extern unsigned char jetsoundend;
extern unsigned char crashsound;
extern unsigned char crashsoundend;
extern unsigned char jaguarsound;
extern unsigned char jaguarsoundend;
extern unsigned char electricsound;
extern unsigned char electricsoundend;
extern unsigned char blanksound;
extern unsigned char blanksoundend;

/* Game tracks data */
extern unsigned int tracks_number;
extern CTrackInfo tracks;

/* Functions */
extern int MenuLoop();
extern void GameLoop(int);
extern void Outtro();

#endif /* __GAME_H__ */