    .include  "jaguar.inc"
    .include  "tiles.inc"
    .include "graphics.inc"


; -------------------------------------------------------
; Globals
; -------------------------------------------------------    
   
    .globl gpusem
    
    ; DrawTunnelProc params
    .globl gpuscrollmapbase 
    .globl gpuLUTYoffset
    .globl gupptrvidmem
    
    ; DrawTileLine params
    .globl gputilemapptr
    .globl gputilesptr
    .globl _gpuscrolltileline
    

; -------------------------------------------------------
; Constants
; -------------------------------------------------------    

BORDERPIXW .equ  16
TUNNELCENTER .equ 64
    
; -------------------------------------------------------
; Externals
; -------------------------------------------------------
    .extern _systemtype
    .extern _lutdata
    .extern _lutleftdata
    .extern _lutrightdata
    .extern _pallutdata
    .extern _pallutleftdata
    .extern _pallutrightdata
    .extern _scrollmap       
    .extern _masktunnel
          
		.gpu
gpucode::
		.org G_RAM	
startgpucode::
        
; -------------------------------
; Dispatch requests from host cpu
gpuinit::
    movei #stackend,r31
    movei #G_FLAGS,r0
    load  (r0),r1
    bset  #4,r1   ; Enable CPU interrupt 
    store r1,(r0)
gpusemcheck:     
    movei #gpusem,r20
    movei #DrawTunnelProc,r21
    movei #DrawTileLineProc,r22
    movei #StopGPU,r23
testgpusem:
    load  (r20),r0
    cmpq  #1,r0       ; Draw tunnel
    jump  eq,(r21)
    nop
    cmpq  #2,r0       ; Draw tileline
    jump  eq,(r22)
    nop
    cmpq  #3,r0       ; Stop GPU
    jump  eq,(r23)
    nop
    jr    testgpusem
    nop


; --------------------------------
;  Draw tunnel
; PARAMS: gpuscrollmapbase (long) ptr to scrollmap base
;         gpuLUTYoffset (long) line to start scanning the lut
rVidmem .equr r16
rLUT  .equr r17
rG_HIDATA .equr r19
rScollmapBase .equr r14
rYcount .equr r7
rXcount .equr r6
rMaskLoOffset .equr r4
rLoopJ  .equr r21
rLoopI  .equr r20

DrawTunnelProc:
    movei #_systemtype,r0
    load  (r0),r0
    movei #60+TUNNELCENTER+16,r2
    movei #maskpos,r3
    cmpq  #1,r0       ; NTSC system?
    jr    eq,.ntscvalues
    nop   
    store r2,(r3)
    movei #_pallutdata,rLUT  
    movei #TUNNELHPAL-8-1,rYcount    
    jr    .continue
    nop  
.ntscvalues:    
    movei #60+TUNNELCENTER,r0
    movei #maskpos,r1
    store r0,(r1)
    movei #_lutdata,rLUT
    movei #TUNNELH-8-1,rYcount
.continue:    
    movei #gpuLUTYoffset,r1    
    load  (r1),r1
    movei #(TUNNELW-(BORDERPIXW*4))*2,r0
    mult  r0,r1
    add   r1,rLUT
    
    movei #gupptrvidmem,rVidmem
    movei #TUNNELW*8+(BORDERPIXW*2),r0
    load  (rVidmem),rVidmem           
    add  r0,rVidmem ; Skip 8 first lines    
    movei #gpuscrollmapbase,r0
    load (r0),rScollmapBase
    movei #.loopi,rLoopI
    movei #.loopj,rLoopJ
    movei #G_HIDATA,rG_HIDATA    
    movei #$FFFF,rMaskLoOffset ; offset mask
.loopj:
    movei #(TUNNELW-(BORDERPIXW*4))/8-1,rXcount

; Preload 1st 4 LUT data
     loadp (rLUT),r0
     
     move  r0,r1    
     load  (rG_HIDATA),r2
     move  r2,r3
.loopi:      
      and   rMaskLoOffset,r0         
      shrq  #16,r1 ; Scale offsets     
      add   rScollmapBase,r0 ; Add base pointer      
      and   rMaskLoOffset,r1  ; Mask offsets
               
      ; Read pixel data  
      loadb (r0),r8 ; read pixel from scrollmap
      and   rMaskLoOffset,r2      
      shrq  #16,r3
      add   rScollmapBase,r1
      and   rMaskLoOffset,r3
      addqt #8,rLUT            
      loadb (r1),r9
      add   rScollmapBase,r2            
      loadb (r2),r10
      add   rScollmapBase,r3 
      
      loadb (r3),r11      
      shlq  #8,r9
      ; Pack pixel data                  
      loadp (rLUT),r0  ; LUT for 4 next pixels
      shlq  #16,r10
      or    r9,r8      
      shlq  #24,r11          
      or    r10,r8
      or    r11,r8     
      move  r0,r1
      load  (rG_HIDATA),r2
                  
      and   rMaskLoOffset,r0     
      shrq  #16,r1    ; Scale offsets                 
      add   rScollmapBase,r0 ; Add base pointer
         
      ; Read pixel data  
      loadb (r0),r12
      and   rMaskLoOffset,r1   ; Mask offsets
      move  r2,r3
      add   rScollmapBase,r1
      and   rMaskLoOffset,r2
      shrq  #16,r3                 
      loadb (r1),r9    
      add   rScollmapBase,r2
      and   rMaskLoOffset,r3
      loadb (r2),r10
      add   rScollmapBase,r3      
      addqt #8,rLUT      
      loadb (r3),r11
      shlq  #8,r9      
      loadp (rLUT),r0  ; LUT load for next loop     
      shlq  #16,r10     ; Pack pixel data           
      or    r9,r12
      shlq  #24,r11         
      or    r10,r12
      or    r11,r12   
      move  r0,r1    
      load  (rG_HIDATA),r2  ; read LUT HIDATA for next loop
      
      store r8,(rG_HIDATA)   ; Store 4 first pixels to write in HIDATA
      move  r2,r3              
      subq  #1,rXcount     ; Decrease X counter      
      storep r12,(rVidmem)  ; r12 contains 4 second pixel to write.          
      jump  pl,(rLoopI)  ; Test X counter and jump
      addqt #8,rVidmem
         
    subq  #1,rYcount
    addqt #BORDERPIXW*2,rVidmem 
    jump pl,(rLoopJ) ; .loopj:
    addqt #BORDERPIXW*2,rVidmem
       
DrawLeftBorder:
    movei #_systemtype,r0
    load  (r0),r0
    cmpq  #1,r0       ; NTSC system?
    jr    eq,.ntscvalues
    nop
    movei #_pallutleftdata,rLUT    ; PAL values
    movei #TUNNELHPAL-8-1,rYcount
    jr    .continue
    nop 
.ntscvalues:
    movei #_lutleftdata,rLUT    ; NTSC values
    movei #TUNNELH-8-1,rYcount    
.continue:
    movei #gpuLUTYoffset,r1
    movei #TUNNELW-(BORDERPIXW*2),r12 ; vidmem jump
    moveq #0,r11      ; vidmem write start (left)
    movei #DrawBorder,r20 
    movei #DrawRightBorder,r27
    jump  (r20)
    nop
    
DrawRightBorder:
    movei #_systemtype,r0
    load  (r0),r0
    cmpq  #1,r0       ; NTSC system?
    jr    eq,.ntscvalues
    nop
    movei #_pallutrightdata,rLUT    ; PAL values
    movei #TUNNELHPAL-8-1,rYcount 
    jr    .continue
    nop
.ntscvalues:
    movei #_lutrightdata,rLUT
    movei #TUNNELH-8-1,rYcount
.continue:
    movei #gpuLUTYoffset,r1
    
    movei #TUNNELW-(BORDERPIXW*2),r12 ; vidmem jump
    move  r12,r11     ; ; vidmem write start (right
     
    movei #MaskTunnel,r27
        
DrawBorder:     
    load  (r1),r1
    movei #BORDERPIXW*2,r0
    mult  r0,r1
    add   r1,rLUT
  
    movei #gupptrvidmem,rVidmem
    movei #TUNNELW*8,r0
    load  (rVidmem),rVidmem    
    add  r0,rVidmem ; Skip 8 first lines  
    add  r11,rVidmem  
    movei #gpuscrollmapbase,r0
    load (r0),rScollmapBase
    movei #.loopi,rLoopI
    movei #.loopj,rLoopJ
    movei #G_HIDATA,rG_HIDATA   
    movei #$FFFF,rMaskLoOffset ; offset mask
.loopj:
    movei #(BORDERPIXW/4)-1,rXcount

; Preload 4 LUT data
     loadp (rLUT),r0
     
     move  r0,r1    
     load  (rG_HIDATA),r2
     move  r2,r3
.loopi:                         
      and   rMaskLoOffset,r0
      shrq  #16,r1 ; Scale offsets                    
      add   rScollmapBase,r0 ; Add base pointer      
            
      ; Read pixel data  
      loadb (r0),r8 ; read pixel from scrollmap
      and   rMaskLoOffset,r1  ; Mask offsets
      addqt #8,rLUT
      add   rScollmapBase,r1
      and   rMaskLoOffset,r2
      shrq  #16,r3                        
      loadb (r1),r9
      and   rMaskLoOffset,r3      
      add   rScollmapBase,r2      
      loadb (r2),r10
      add   rScollmapBase,r3      
      loadb (r3),r11
      
      ; Pack pixel data                               
      shlq  #16,r9
      loadp (rLUT),r0 ; Read LUT data for next loop
      shlq  #16,r11
      or  r9,r8
      or  r11,r10
      move  r8,r9
      move  r10,r11
      shlq  #8,r9
      shlq  #8,r11
      or  r9,r8      
      or  r11,r10
      move  r0,r1      
      load  (rG_HIDATA),r2
      store r10,(rG_HIDATA) ; HIDATA contains the 4 first pixel to write
      move  r2,r3    
        
      subq  #1,rXcount     ; Decrease X counter   
      storep r8,(rVidmem)  ; r8 contains 4 second pixel to write.       
      jump  pl,(rLoopI)  ; Test X counter and jump
      addqt #8,rVidmem
   
    add r12,rVidmem    
      
    subq  #1,rYcount 
    jump pl,(rLoopJ) ; .loopj:
    nop  
         
    jump  (r27)
    nop
      
MaskTunnel:
    movei #B_PATD,r10
    movei #B_PATD+4,r13
    movei #A1_PIXEL,r11
    movei #A2_PIXEL,r12    
    movei #A1_STEP,r14
    movei #A1_BASE,r15
    movei #A2_BASE,r16
    
    moveq #0,r0     ; PIXEL
    movei #-16,r1    
    movei #$FFFF,r3
    and   r3,r1
    movei #(1<<16),r3
    or    r3,r1   ; A1 STEP
    store r0,(r11)  ; Store A1_PIXEL
    store r0,(r12)  ; Store A2_PIXEL
    store r1,(r14)  ; Store A1_STEP
    store r0,(r10)  ; BPATD1
    store r0,(r13)  ; BPATD2
             
    movei #gpuLUTYoffset,r4
    movei #maskpos,r3
    load  (r4),r4
    load  (r3),r3     
    movei #TUNNELW,r1
    sub   r4,r3
    movei #(TUNNELW-16)/2,r2
    mult  r3,r1   ; Y*TUNNELW
    movei #gupptrvidmem,r4
    load  (r4),r4
    add   r1,r2   ; Y*TUNNELW + (TUNNELW/2)-8
    add   r2,r4  ; A1 BASE
    movei #_masktunnel+512,r5 ; A2 BASE   
    movei #(16<<16)|16,r2 ; COUNT
    store r4,(r15)  ; Store A1 BASE
    store r5,(r16)  ; Store A2 BASE
    
    movei #B_CMD,r10
    movei #A1_FLAGS,r11
    movei #A2_FLAGS,r12
    movei #B_COUNT,r13
    movei #PITCH1|PIXEL8|WID320|XADDPHR,r0
    movei #PITCH1|PIXEL8|WID16|XADDPHR,r1
    movei #SRCEN|DSTEN|LFU_REPLACE|DCOMPEN|UPDA1,r3
    store r0,(r11)
    store r1,(r12)
    store r2,(r13)  ; BCOUNT
    store r3,(r10)  ; BCMD
.test_blit:
      load  (r10),r0
      btst  #0,r0
      jr  eq,.test_blit:
      nop    
        
    movei #gpusem,r2
    moveq #31,r0      ; Were are done
    store r0,(r2)

    movei #G_CTRL,r1
    load  (r1),r0
    bset  #1,r0   ; CPU INT    
    store r0,(r1)   ; trigger CPU INT
    
    movei #gpusemcheck,r20
    jump  (r20)
    nop
    
  .equrundef  rG_HIDATA
  .equrundef  rLUT
  .equrundef  rVidmem
  .equrundef  rScollmapBase
  .equrundef  rYcount
  .equrundef  rXcount
  .equrundef  rMaskHiOffset
  .equrundef  rMaskLoOffset
  .equrundef  rLoopJ
  .equrundef  rLoopI
  
  
; ------------------------------------------------
; DrawTileLineProc
; PARAMS: gputilemapptr (long) ptr to tilemap
;         gpulinenumber (long) line number
rB_CMD    .equr r14
rB_COUNT  .equr r15
rA1_PIXEL .equr r16
rA2_PIXEL .equr r17
rA1_BASE  .equr r18
rA2_BASE  .equr r19

DrawTileLineProc:
      ; Load registers
      movei #B_CMD,rB_CMD
      movei #B_COUNT,rB_COUNT
      movei #A1_PIXEL,rA1_PIXEL
      movei #A2_PIXEL,rA2_PIXEL
      movei #A1_BASE,rA1_BASE
      movei #A2_BASE,rA2_BASE

; *** Duplicate last tile line 
;   If we move back to the top of the scrollmap
;     we must copy the last tile line to the high buffer
      movei #_gpuscrolltileline,r0      
      load  (r0),r2
      movei #DontCopyLastLine,r20      
      addqt #1,r2
      moveq #16,r1
      store r2,(r0)
      cmp r1,r2
      jump  mi,(r20)
      nop
      moveq #8,r1
      store r1,(r0)
      movei #_scrollmap,r13 ; scrollmap address
      movei #A2_FLAGS,r10
      movei #A1_FLAGS,r11      
      movei #(SCROLLMAP_H-TILE_H)<<16,r0 ; Destination pixel
      movei #(SCROLLMAP_H*2-TILE_H)<<16,r1 ; Source pixel
    ;  movei #(TILE_H<<16)|(SCROLLMAP_W),r2 ; Count
      movei #(TILE_H<<16)|(64),r2 ; Count
    ;  movei #PITCH1|PIXEL8|SCROLLMAPBW|XADDPHR,r3 ; Blitter flags      
      movei #PITCH1|PIXEL32|WID64|XADDPHR,r3 ; Blitter flags
      movei #SRCEN|LFU_REPLACE,r5
      
      store r3,(r10)  ; A1 flags
      store r3,(r11)  ; A2 flags
      store r13,(rA1_BASE)
      store r13,(rA2_BASE)
      store r0,(rA1_PIXEL)
      store r1,(rA2_PIXEL)      
      store r2,(rB_COUNT)
      store r5,(rB_CMD)
.test_blit:
      load  (rB_CMD),r0
      btst  #0,r0
      jr  eq,.test_blit:
      nop
DontCopyLastLine:

; *** Draw tile line     
;  Draw current tile line 
      ; Compute step      
   ;  movei #TILE_W,r0
      movei #8,r0
      movei #$FFFF,r1
      movei #1<<16,r2
      movei #A1_STEP,r10
      neg   r0
      and   r1,r0       
      or    r2,r0
      store r0,(r10)

      ; Set blitter flags
   ;   movei #PITCH1|PIXEL8|TILE_BW|XADDPHR,r0
   ;   movei #PITCH1|PIXEL8|SCROLLMAPBW|XADDPHR,r1
      movei #PITCH1|PIXEL32|WID8|XADDPHR,r0
      movei #PITCH1|PIXEL32|WID64|XADDPHR,r1
      movei #A2_FLAGS,r10
      movei #A1_FLAGS,r11        
      store r0,(r10)
      store r1,(r11)
      
      ; Save count
  ;    movei #(TILE_H<<16)|(TILE_W),r2 
      movei #(TILE_H<<16)|(8),r2
      
      movei #_gpuscrolltileline,r1
      load  (r1),r1
      movei #_scrollmap,r13 ; scrollmap address
      shlq  #TILE_SHH,r1    ; TILE Y position in scrollmap
      shlq  #SCROLLMAP_SW,r1 ; TILE offset in scrollmap    
      add   r1,r13        ; TILE address
            
      moveq #0,r3
      movei #gputilemapptr,r11
      movei #gputilesptr,r10      
      load  (r11),r11
      load  (r10),r10
            
      movei #SRCEN|LFU_REPLACE|UPDA1,r5
      movei #TILE_W,r6
        
      movei #.loop,r20      
      moveq #N_TILES_X-1,r4    
.loop:         
      loadb (r11),r0 ; tile selected
      move  r10,r12      
      shlq  #TILE_SHW,r0    
      shlq  #TILE_SHH,r0
      addq  #1,r11      
      add   r0,r12 ; ptr to tile bitmap

      store r3,(rA2_PIXEL)
      store r3,(rA1_PIXEL)                  
      store r12,(rA2_BASE)
      store r13,(rA1_BASE)
      store r2,(rB_COUNT)      
      store r5,(rB_CMD) 

.testblit:
      load  (rB_CMD),r1
      btst  #0,r1
      jr    eq,.testblit
      nop
                 
      subq  #1,r4
      jump  pl,(r20)
      add r6,r13
      
; *** Duplicate tile line in low buffer to high buffer
DuplicateTileLine:
      movei #_gpuscrolltileline,r1
      load  (r1),r1
      movei #_scrollmap,r13 ; scrollmap address
      subq  #1,r1     ; (tileline-1)
      movei #SCROLLMAP_H,r2
      shlq  #TILE_SHH,r1  ; line*TILE_H
      movei #DontCopyCurrentLine,r20
      move  r1,r3   ; Save (line-1)*TILE_H
      sub   r2,r1   ; duplicateline = (line-1)*TILE_H - SCROLLMAP_H
      jump  mi,(r20)
      nop
      
      movei #SRCEN|LFU_REPLACE,r5
      shlq  #SCROLLMAP_SW,r3  ; offset source      
   ;   movei #PITCH1|PIXEL8|SCROLLMAPBW|XADDPHR,r0      
      movei #PITCH1|PIXEL32|WID64|XADDPHR,r0
      movei #A2_FLAGS,r10
      movei #A1_FLAGS,r11            
      store r0,(r10)
      store r0,(r11)     
  ;    movei #(TILE_H<<16)|SCROLLMAP_W,r2 ;  ; Pixel count
       movei #(TILE_H<<16)|(64),r2 ;  ; Pixel count
            
      shlq  #16,r1  ; duplicate line => Y pixel | X=0 for A1_PIXEL
      store r13,(rA1_BASE);
      store r1,(rA1_PIXEL)      
      moveq #0,r0 
      
      add   r3,r13
      store r0,(rA2_PIXEL)      
      store r13,(rA2_BASE)
      store r2,(rB_COUNT)
      store r5,(rB_CMD)
.testblit:
      load  (rB_CMD),r1
      btst  #0,r1
      jr    eq,.testblit
      nop
DontCopyCurrentLine:      
    movei #gpusem,r20
    moveq #31,r0      ; Were are done
    store r0,(r20)

    movei #G_CTRL,r1
    load  (r1),r0
    bset  #1,r0   ; CPU INT    
    store r0,(r1)   ; trigger CPU INT
    movei #gpusemcheck,r20
    jump  (r20)
    nop

  .equrundef rB_CMD
  .equrundef rB_COUNT
  .equrundef rA1_PIXEL
  .equrundef rA2_PIXEL
  .equrundef rA1_BASE
  .equrundef rA2_BASE
  
      
; ------------------------------------------------
;  StopGPU
; No parameters        
StopGPU:	
	movei  #0,r0		; Stop GPU
	movei  #G_CTRL,r1
	store   r0,(r1)	 
		nop             		; Two "feet" on the brake pedal
		nop
	
    .long
gpusem: dc.l  0
gpuscrollmapbase:  dc.l 0
gpuLUTYoffset:    dc.l  0
gupptrvidmem:     dc.l  0

gputilemapptr:  dc.l 0
gputilesptr:   dc.l  0
_gpuscrolltileline:  dc.l 0

maskpos:    dc.l  0

stackbegin: dc.l 0  
            dc.l 0
            dc.l 0
stackend:   dc.l 0

endgpucode::
	  nop
		nop


